"""Ampio CAN protocol implementation."""

import asyncio
import logging
import struct
from enum import Enum, unique


_LOG = logging.getLogger(__name__)


@unique
class Type(Enum):
    """This is a frame type enum for USB CAN module frames."""

    # send raw can frame
    CAN = 0x00
    SEND_VALUE_WITH_MASK = 0x10
    SEND_VALUE_WITH_INDEX = 0x11


class CanType(Enum):
    """CAN frame types."""

    BROADCAST = 0xfe
    DATA_REPLY = 0x0f


class AmpioCanProtocol(asyncio.Protocol):
    """This is an Ampio CAN protocol implementation for serial USB module."""

    def __init__(self,
                 on_connected=None,
                 on_can_broadcast_received=None,
                 on_can_data_received=None,
                 is_relevant=None):
        """Initialize the :class: pyampio.can.AmpioCanProtocol object.

        Args:
            on_connected (callable): The callback function for on_connected event
            on_can_broadcast_received (callable): The callback function for on_broadcast_received event
            on_can_data_received (callable): The callback function for on can_data_received event
            is_relevant (callable): Filter function to filter relevant frames. i.e. can_id based.

        """
        if is_relevant:
            self._is_relevant = is_relevant
        else:
            self._is_relevant = lambda can_id: True

        self._assembly_buffer = b''
        self.transport = None
        # on connected coroutine
        self._on_connected = on_connected
        # 0xfe
        self._on_can_broadcast_received = on_can_broadcast_received
        # 0x0f
        self._on_can_data_received = on_can_data_received

        self.frame_type_map = {
            CanType.BROADCAST: self._on_can_broadcast_received,
            CanType.DATA_REPLY: self._on_can_data_received,
        }

    def connection_made(self, transport):
        """Protocol connection made."""
        self.transport = transport
        _LOG.debug('port opened')
        self.transport.serial.rts = True
        self.transport.serial.dtr = True
        self.transport.serial.write_timeout = 10
        self.transport.serial.reset_input_buffer()
        self.transport.serial.reset_output_buffer()
        if self._on_connected:
            self.transport._loop.create_task(self._on_connected(self))

    @asyncio.coroutine
    def open_connection(self):
        """Protocol connection open."""
        _LOG.debug("Opening connection")

    def send_api_function(self, api_func, can_id, data):
        """Send API function to module.

        Args:
            api_func (int): API function
            can_id (int): CAN ID
            data (bytearray): Data

        """
        self.send_frame(api_func, can_id, bytearray(data))

    def send_module_command(self, func, gateway_can_id, can_id=None):
        """Send module command.

        Args:
            own_can_id (int): own can Id
            can_id (int): destination module can id
            func (int): module function number

        """
        data = can_id.to_bytes(4, byteorder='big') + func.to_bytes(1, byteorder='big') if can_id else \
            func.to_bytes(1, byteorder='big')
        self.send_can_frame(gateway_can_id, data)

    def send_can_frame(self, can_id, data):
        """Send the CAN frame.

        Args:
            can_id (int): CAN ID
            data (bytearray): Array of CAN data bytes

        """
        self.send_frame(Type.CAN, can_id, data)

    def send_frame(self, frame_type, can_id, data):
        """Send the RAW frame to serial interface.

        Args:
            frame_type (Type): Frame type enum
            can_id (int): CAN ID
            data (bytearray): Array of CAN data bytes

        """
        # encode can_id to bytes
        can_id_bytes = can_id.to_bytes(4, byteorder='big')
        # build the frame
        frame = bytearray(b'\x2d\xd4\x00\x00' + can_id_bytes + data + b'\x00')
        # calculate and update length
        frame[2] = len(frame) - 3
        # update frame type
        frame[3] = frame_type.value
        # calculate CRC
        frame[-1] = sum(frame[:-1]) & 0xff
        self.transport.write(frame)
        can_data_str = " ".join(["{:02x}".format(c) for c in frame])
        _LOG.debug("CAN SERIAL OUT: frame=[{}]".format(can_data_str))

    @asyncio.coroutine
    def _process_frame(self, frame):
        frame_len = frame[0]
        if frame_len == 0:
            _LOG.warning("Frame length is zero")
            return
        try:
            frame_type = Type(frame[1])
        except ValueError:
            _LOG.warning("Unknown frame type {}".format(frame[1]))
            return

        if frame_type == Type.CAN:
            can_len = frame[0] - 6
            if not (0 < can_len < 9):
                _LOG.error("Invalid CAN frame length: {}".format(can_len))
                return

            try:
                device_id = struct.unpack(">L", frame[2:6])[0]
            except struct.error:
                _LOG.error("Unable to decode CAN ID: {}".format(frame))

            can_data = frame[6:]
            yield from self._process_can_frame(device_id, can_len, can_data)

    @asyncio.coroutine
    def _process_can_frame(self, can_id, can_len, can_data):
        if self._is_relevant(can_id):
            can_data_str = " ".join(["{:02x}".format(c) for c in can_data])
            _LOG.debug("CAN IN: id={:08x} len={:02x} data=[{}]".format(can_id, can_len, can_data_str))

        try:
            coro = self.frame_type_map[CanType(can_data[0])]
            if coro:
                yield from coro(can_id, can_data)
        except ValueError:
            can_data_str = " ".join(["{:02x}".format(c) for c in can_data])
            _LOG.warning("Frame type not implemented: id={:08x} len={:02x} data=[{}]".format(
                can_id, can_len, can_data_str))

    def data_received(self, data):
        """Process received raw data from the wire."""
        # _LOG.debug("-" * 20)
        # _LOG.debug('Data received: {!r}'.format(data.hex()))
        # _LOG.debug('Assembly buffer: {!r}'.format(self._assembly_buffer.hex()))
        # _LOG.debug("-" * 20)
        data = self._assembly_buffer + data
        data_length = len(data)
        while data_length > 3:
            if data[0:2] == b'\x2d\xd4':
                frame_length = data[2] + 3
                if frame_length > data_length:
                    # frame fragment
                    break
                frame = data[:frame_length]
                crc = sum(frame[:-1]) & 0xff
                if crc == frame[-1]:
                    # _LOG.debug("Data: {!r}".format(frame.hex()))
                    frame = frame[2:-1]
                    self.transport._loop.create_task(self._process_frame(frame))
                else:
                    _LOG.error("CRC error: {!r} calc crc={:02x} frame crc={:02x}".format(frame.hex(), crc, frame[-1]))

                data = data[frame_length:]
                data_length -= frame_length
            else:
                # walk through the buffer trying to find preamble 0x2dd4
                _LOG.warning("Synchronization error. Trying to find a frame preamble")
                _LOG.warning(data)
                data = data[1:]
                data_length -= 1
                # anyway can't assemble frame - resetting the assembly buffer
                self._assembly_buffer = b''

        self._assembly_buffer = data

    def connection_lost(self, exc):
        """Handle the connection lost event."""
        _LOG.error("Port closed")
        self.transport._loop.stop()

    def pause_writing(self):
        """Handle pause writing event."""
        _LOG.warning('pause writing')
        _LOG.warning(self.transport.get_write_buffer_size())

    def pause_reading(self):
        """Handle pause reading event."""
        _LOG.warning(self.transport.get_write_buffer_size())
        _LOG.warning('resume writing')



"""This is a main AmpioGateway implementation module."""

import asyncio
import logging
from serial_asyncio import create_serial_connection
from enum import Enum, IntEnum
from pyampio.can import AmpioCanProtocol, Type
from pyampio.modules import AmpioModules


_LOG = logging.getLogger(__name__)


class GatewayState(Enum):
    """This is an enum class for Gateway State."""

    INIT = 0
    QUERY_MODULES = 4
    QUERY_MODULE_NAME = 5
    QUERY_MODULE_DETAILS = 6
    QUERY_MODULE_ITEM_NAMES = 7
    READY = 8


GATEWAY_CAN_ID = 0xdeadbeef

# Timeout for modules query after moving to next stage
QUERY_MODULES_TIMEOUT = 2


class ModuleFunction(IntEnum):
    """This is an enmu class for Module Functions."""

    Discovery = 0xff
    QueryNames = 0x01
    QueryDetails = 0x03
    QueryItemNames = 0x08


class AmpioGateway:
    """This is a AmpioGateway class handling the connection USB<->Ampio CAN bus."""

    def __init__(self, port=None, loop=None, no_query_module_details=False):
        """Initialize the :class: pyampio.AmpioGateway object.

        Args:
            port (str): The USB port path i.e. /dev/usb1
            loop: Asynction Event Loop

        """
        self._port = port
        self._loop = loop
        self._no_query_module_details = no_query_module_details

        self._modules = AmpioModules()
        self._on_discovered_callbacks = []

        self._is_relevant = lambda can_id: True
        self._protocol_coro = create_serial_connection(
            loop, lambda: AmpioCanProtocol(
                on_connected=self.on_connected,
                on_can_broadcast_received=self.on_can_broadcast_received,
                on_can_data_received=self.on_can_data_received,
                is_relevant=self._is_relevant
            ), port, baudrate=115200
        )
        self.protocol = None
        self._state = GatewayState.INIT
        asyncio.ensure_future(self._protocol_coro, loop=self._loop)

    @property
    def state(self):
        """Return the current object state.

        Returns:
            GatewayState: The current gateway state.

        """
        return self._state

    @property
    def is_ready(self):
        """Return True if gateway is connected and Ready."""
        return self._state == GatewayState.READY

    @property
    def is_connected(self):
        """Return True if gateway is connected to serial port."""
        return self.protocol is not None

    @state.setter
    def state(self, value):
        """Set the object state."""
        _LOG.info("State change {}->{}".format(self._state.name, value.name))
        self._state = value

    def add_on_discovered_callback(self, callback):
        """Add the callback for On Discovered Event."""
        if callback not in self._on_discovered_callbacks:
            self._on_discovered_callbacks.append(callback)

    def remove_on_discovered_callback(self, callback):
        """Remove the callback for On Discovered Event."""
        try:
            self._on_discovered_callbacks.remove(callback)
        except ValueError:
            pass

    @asyncio.coroutine
    def on_discovered(self):
        """Fire the callback when all modules are fully discovered."""
        for callback in self._on_discovered_callbacks:
            yield from callback(modules=self._modules)

    @asyncio.coroutine
    def on_connected(self, protocol):
        """Call on gateway connected event."""
        _LOG.debug("Connected")
        self.protocol = protocol
        self.state = GatewayState.QUERY_MODULES
        # send the module discovery
        self.protocol.send_module_command(ModuleFunction.Discovery, GATEWAY_CAN_ID)
        self._loop.call_later(QUERY_MODULES_TIMEOUT, self.discovery_finished)

    def discovery_finished(self):
        """Schedule the Query Module Name phase when all module discovered."""
        module_quantity = len(self._modules)
        _LOG.info("Total number of modules discovered: {}".format(module_quantity))

        self.state = GatewayState.QUERY_MODULE_NAME
        self._loop.create_task(self.query_module_names())

    @asyncio.coroutine
    def query_module_names(self):
        """Run the Query Module Names phase."""
        for can_id, mod in self._modules:
            self.protocol.send_module_command(ModuleFunction.QueryNames, GATEWAY_CAN_ID, can_id)
            yield from self._modules.is_name_updated(can_id)

        _LOG.info("Module names discovered")
        if self._no_query_module_details:
            self.state = GatewayState.READY
            self._loop.create_task(self.on_discovered())
        else:
            self.state = GatewayState.QUERY_MODULE_DETAILS
            self._loop.create_task(self.query_module_details())

    @asyncio.coroutine
    def query_module_details(self):
        """Run the module details discovery phase."""
        for can_id, mod in self._modules:
            self.protocol.send_module_command(ModuleFunction.QueryDetails, GATEWAY_CAN_ID, can_id)
            yield from self._modules.is_details_updated(can_id)
        for can_id, mod in self._modules:
            _LOG.info(mod)
        _LOG.info("Module details discovered")
        self.state = GatewayState.READY
        self._loop.create_task(self.on_discovered())

        # TODO(klstanie): Add the attribute names discovery logic
        # Attribute names discovery not implemented yet
        # self.state = GatewayState.ATTRIBUTE_NAMES_DISCOVERY
        # self._loop.create_task(self.attribute_names_discovery())

    @asyncio.coroutine
    def query(self):
        """Run the attribute names discovery phase."""
        for can_id, mod in dict(self._modules.modules).items():
            self.protocol.send_module_command(ModuleFunction.QueryItemNames, GATEWAY_CAN_ID, can_id)
            yield from self._modules.is_attribute_names_updated(can_id)
        for can_id, mod in self._modules.modules.items():
            _LOG.info(mod)
        _LOG.info("Attribute names discovered")

    @asyncio.coroutine
    def close(self):
        """Close the Gateway Serial Connection."""
        self._loop.call_later(1, self.protocol.transport.close)

    @asyncio.coroutine
    def on_can_broadcast_received(self, can_id, can_data):
        """Handle the broadcast frame."""
        if self.state == GatewayState.READY:
            can_data_str = " ".join(["{:02x}".format(c) for c in can_data])
            _LOG.debug("CAN IN: id={:08x} data=[{}]".format(can_id, can_data_str))
            self._modules.broadcast_received(can_id, can_data)

    @asyncio.coroutine
    def on_can_data_received(self, can_id, can_data):
        """Handle the data frame."""
        if self.state == GatewayState.QUERY_MODULES:
            self._modules.add_module(can_id, can_data)
        elif self.state == GatewayState.QUERY_MODULE_NAME:
            self._modules.update_name(can_id, can_data)
        elif self.state == GatewayState.QUERY_MODULE_DETAILS:
            self._modules.update_details(can_id, can_data)

    def register_on_value_change_callback(self, can_id, attribute, index, callback):
        """Register the value changed callback."""
        self._modules.register_on_value_changed_callback(can_id, attribute, index, callback)

    def get_item_state(self, can_id, attribute, index):
        """Return the item state from ampio module."""
        return self._modules.get_state(can_id, attribute, index)

    def get_item_last_state(self, can_id, attribute, index):
        """Return the item last state from ampio module."""
        return self._modules.get_last_state(can_id, attribute, index)

    def get_item_measurement_unit(self, can_id, attribute, index):
        """Return the item measurement unit."""
        return self._modules.get_measurement_unit(can_id, attribute, index)

    def get_module_name(self, can_id):
        """Return Ampio module name."""
        return self._modules.get_module_name(can_id)

    def get_module_part_number(self, can_id):
        """Return Ampio module part number."""
        return self._modules.get_module_part_number(can_id)

    @asyncio.coroutine
    def send_value_with_mask(self, can_id, mask, value):
        """Send value to module using item mask.

        Params:
            can_id (int): Module CAN ID
            mask(int): 2 byte item mask
            value (int): Single byte value

        """
        if self.protocol:
            value = value.to_bytes(1, byteorder='big')
            mask &= 0xffff
            mask = mask.to_bytes(2, byteorder='big')
            self.protocol.send_frame(Type.SEND_VALUE_WITH_MASK, can_id, value + mask)
        pass

    @asyncio.coroutine
    def send_value_with_index(self, can_id, index, value):
        """Send value to module using item index.

        Params:
            can_id (int): Module CAN ID
            index (int): 1-based item index
            value (int): Single byte value

        """
        if self.protocol:
            value = value.to_bytes(1, byteorder='big')
            index -= 1
            index &= 0xff
            index = index.to_bytes(1, byteorder='big')
            self.protocol.send_frame(Type.SEND_VALUE_WITH_INDEX, can_id, value + index)



#!/usr/bin/env python
"""The ampio command line implementation."""

import asyncio
from pyampio.gateway import AmpioGateway
import pyampio.version as v
import logging


try:
    import click
except ImportError:
    print("Install click python package\n pip install click")
    exit()


class BasedIntParamType(click.ParamType):
    """Type for based int."""

    name = 'integer'

    def convert(self, value, param, ctx):
        """Convert value from hex, oct, dec to dec."""
        try:
            if value[:2].lower() == '0x':
                return int(value[2:], 16)
            elif value[:1] == '0':
                return int(value, 8)
            return int(value, 10)
        except ValueError:
            self.fail('%s is not a valid integer' % value, param, ctx)


BASED_INT = BasedIntParamType()


def on_value_changed(modules, can_id, attribute, index, old_value, new_value, unit):
    """Handle the value changed event.

    Args:
        modules (ModuleManager): Module manager object
        can_id (int): CAN ID
        attribute (str): The attribute name i.e. input, bin_input, temperature, flag, etc.
        index (int): The attribute index (1-based)
        old_value: Old value
        new_value: New value
        unit (str): Unit

    """
    mod = modules.get_module(can_id)
    if mod:
        print("{:08x}/{:08x} {:16} {:32} {}({}) changed {}{}->{}{}".format(
            mod.physical_can_id, can_id, mod.part_number, mod.name, attribute, index, old_value, unit, new_value, unit)
        )


# @asyncio.coroutine
# def on_discovered_call(loop, can_id, command, index, value, ampio, modules):
#
#     # yield from ampio.send_value_with_index(0x1ae0, 0x0, 4)
#     yield from ampio.send_value_with_index(0x13ec, 0x00, 5)
#
#     #ampio.protocol.close()
#
#     # loop.stop()

@asyncio.coroutine
def send_command(ampio, can_id, command, index, value):
    """Send command to Ampio Module."""
    while not ampio.is_connected:
        yield
    if command == 'set_value':
        yield from ampio.send_value_with_index(can_id, value, index)
    # yield from ampio.send_value_with_mask(0x1ae0, 0x00, 0x08) # mask 0x08 - index = 4

    # yield from ampio.send_value_with_index(0x13ec, 0xff, 4)
    # yield from ampio.send_value_with_mask(0x13ec, 0x00, 0x10)
    yield from ampio.close()


@asyncio.coroutine
def on_discovered(modules):
    """Handle the on discovered event when discovery phase is finished."""
    for _, mod in modules.modules.items():
        print(mod)
    print("Discovered {} modules.".format(len(modules.modules)))
    print("Registering for value change events.")
    # Subscribe to all values
    for can_id, mod in modules.modules.items():
        modules.add_on_value_changed_callback(can_id=can_id, attribute=None, index=None, callback=on_value_changed)
    # modules.add_on_value_changed_callback(can_id=0x1ecc, attribute='bin_input', index=9, callback=on_value_changed)


log_levels = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "ERROR": logging.ERROR,
    "NONE": logging.NOTSET,
}


@click.group()
def cli():
    """CLI Entry."""
    pass


@cli.command("call", help="Call the API function", short_help="Call API")
@click.option("--port", required=True, envvar='AMPIO_PORT', type=click.Path(),
              help='The USB interface full path i.e. /dev/cu.usbserial-DN01D1W1. '
                   'If no --port option provided the AMPIO_PORT environment variable is used.')
@click.option("--log-level", type=click.Choice(["NONE", "DEBUG", "INFO", "ERROR"]),
              show_default=True, default='ERROR',
              help='Logging level.')
@click.option("--can_id", required=True, type=BASED_INT,
              help="CAN ID")
@click.option("--command", required=True, type=str,
              help="Command to execute: [set_value,]")
@click.option("--index", required=True, type=int,
              help="Value Index")
@click.option("--value", required=True, type=BASED_INT,
              help="Value")
def call(port, log_level, can_id, command, index, value):
    """Call module function."""
    formatter = "[%(asctime)s] %(levelname)s - %(message)s"
    logging.basicConfig(level=log_levels[log_level], format=formatter)

    loop = asyncio.get_event_loop()
    ampio_gw = AmpioGateway(port=port, loop=loop)
    # ampio_gw.add_on_discovered_callback(partial(on_discovered_call, loop, can_id, command, index, value, ampio_gw))
    loop.create_task(send_command(ampio_gw, can_id, command, index, value))
    loop.run_forever()
    loop.close()
    pass


@cli.command("run", help="Run Ampio Gateway", short_help="Run Gateway")
@click.option("--port", required=True, envvar='AMPIO_PORT', type=click.Path(),
              help='The USB interface full path i.e. /dev/cu.usbserial-DN01D1W1. '
                   'If no --port option provided the AMPIO_PORT environment variable is used.')
@click.option("--no-query-details", required=False, is_flag=True,
              help='Run the full module details query')
@click.option("--log-level", type=click.Choice(["NONE", "DEBUG", "INFO", "ERROR"]),
              show_default=True, default='ERROR',
              help='Logging level.')
def run(port, log_level, no_query_details):
    """Run main function."""
    formatter = "[%(asctime)s] %(levelname)s - %(message)s"
    logging.basicConfig(level=log_levels[log_level], format=formatter)

    loop = asyncio.get_event_loop()
    ampio_gw = AmpioGateway(port=port, loop=loop, no_query_module_details=no_query_details)
    ampio_gw.add_on_discovered_callback(on_discovered)
    loop.run_forever()
    loop.close()


@cli.command("version", help="Display the PyAMPIO version", short_help="Version")
def show_version():
    """Display module version number."""
    click.echo("PyAMPIO Version {}\n(c) 2018, Klaudiusz Staniek".format(v.__version__))


if __name__ == '__main__':
    cli()



"""This is a helper module with value conversion functions."""


from calendar import day_name


def convert_temperature_int(value):
    """Convert the integer temperature value to signed."""
    return int(value) - 100


def convert_temperature(value):
    """Convert the decimal dotted temperature value (single digit)."""
    return (int(value) - 1000) / 10


def convert_meteo(value):
    """Convert the value from METEO station."""
    return (int(value) - 10000) / 10


def convert_weekday(value):
    """Convert the week day number to weekday name."""
    try:
        return day_name[value - 1]
    except (TypeError, IndexError):
        return "N/A"



#!/usr/bin/env python

import re
from setuptools import setup, find_packages

from codecs import open
from os import path

pkg_name = 'pyampio'

here = path.abspath(path.dirname(__file__))
version_file = 'version.py'

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(here, pkg_name, 'version.py'), encoding='utf-8') as f:
    data = f.read()

    match = re.search("__version__ = '([^']+)'", data)
    assert match, 'Cannot find version number in {} package'.format(pkg_name)
    version = match.group(1)

setup(name=pkg_name,
      version=version,
      description='A Python Ampio Server USB<->CAN',
      long_description=long_description,
      url='http://github.com/kstaniek/pyampio',
      author='Klaudiusz Staniek',
      author_email='kstaniek@gmail.com',
      license='Apache 2.0',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'Topic :: System :: Hardware :: Hardware Drivers',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3 :: Only',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
      ],
      packages=find_packages(exclude=['tests']),
      keywords='ampio automation can',
      install_requires=['pyserial==3.4', 'pyserial-asyncio==0.4', 'PyYAML==3.12', 'voluptuous==0.10.5', 'click==6.7'],
      include_package_data=True,
      data_files=[('pyampio', ['pyampio/modules.yaml'])],
      zip_safe=False,
      entry_points={
          'console_scripts': [
              'ampio = pyampio.__main__:cli',
          ]
      }
)



"""This is a module implementing the Ampio Module class."""

import yaml
import re
from functools import partial
from collections import defaultdict
import datetime as dt
from pyampio.broadcast import BroadcastCache
from pyampio.broadcast import BroadcastTypes # noqa
from pyampio.converters import convert_temperature, convert_weekday, convert_meteo, convert_temperature_int  # noqa
import logging
from voluptuous import Schema, Required, All, Lower, Any, Optional, Coerce, Invalid

_LOG = logging.getLogger(__name__)

QUERY_MODULE_DETAILS_TIMEOUT_SECONDS = 2


class AmpioModule:
    """This is AmpioModule class representing the Ampio Module."""

    def __init__(self, can_id, code, protocol, software_version, part_number, description, attributes):
        """Initialize the :class: pyampio.modules.AmpioModule object.

        Args:
            can_id (int): CAN_ID
            code (int): The module code numer
            protocol (int): Protocol version
            software_version (int): Software version
            part_number (str): Ampio Module P/N
            description (str): Module description
            attributes (dict): Module attributed

        """
        self.can_id = can_id
        self.physical_can_id = can_id
        self.pcb = 0
        self.code = code
        self.protocol = protocol
        self.software_version = software_version
        self.part_number = part_number
        self.description = description
        self.attributes = attributes
        # (attribute, index) -> (broadcast, index, unit) map
        self.map = {}
        self._broadcasts = {}
        self._broadcast_cache = BroadcastCache(self.on_change)
        # TODO: Remove
        self._callbacks = defaultdict(list)
        self._state_update_callbacks = defaultdict(list)
        self._name_parts = [""] * 6
        self._details_flag = 0x00

        self._make_attribute_map()

    def __str__(self):
        """Return the string representation of the module details."""
        can_id = "         {:08X}".format(self.can_id) if self.physical_can_id == self.can_id \
            else "{:08X}/{:08X}".format(self.physical_can_id, self.can_id)
        return "{} {:2}-{:16} {:02} {:45} {:5d} {:2d} {}".format(
            can_id, self.code, self.part_number, self.protocol, self.description,
            self.software_version, self.pcb, self.name)
    __repr__ = __str__

    @property
    def name(self):
        """Return the discovered module name."""
        return ("".join(self._name_parts)).strip()

    @property
    def is_name_complete(self):
        """Return true if module name discovery is completed."""
        return "" not in self._name_parts

    @property
    def is_details_complete(self):
        """Return true if module details discovery is completed."""
        return self._details_flag == 0xff

    def update_pcb_version(self, pcb):
        """Update the pcb version."""
        self.pcb = pcb

    def broadcast_received(self, can_data):
        """Handle the received broadcast frame."""
        self._broadcast_cache.update(can_data[1:])

    def update_name(self, index, data):
        """Handle the frame from the block with module name."""
        self._name_parts[index] = data.decode('cp1250')

    def update_details(self, can_data, update_can_id_callback):
        """Handle the frame with module details."""
        data_id = can_data[1]
        if data_id == 0x00:
            data_h = can_data[2]
            data_l = can_data[3]
            uptime = (can_data[4] << 16) | (can_data[5] << 8) | can_data[6]
            packet_size = can_data[7]
            _LOG.debug("UPDATE DETAILS 0x00: {:08x} data_h={:02x} data_l={:02x} uptime={} packet_size={}".format(
                self.can_id, data_h, data_l, uptime, packet_size))
            self._details_flag |= 0x01

        elif data_id == 0x01:
            zd_counter = (can_data[2] << 16) | (can_data[3] << 8) | can_data[4]
            reset_counter = (can_data[5] << 8) | can_data[6]
            can_errors = can_data[7]
            _LOG.debug("UPDATE DETAILS 0x01: {:08x} zd_counter={:06x} reset_counter={:04x} can_errors={}".format(
                self.can_id, zd_counter, reset_counter, can_errors))
            self._details_flag |= 0x02

        elif data_id == 0x02:
            com_stat = can_data[2]
            rx_errcnt = can_data[3]
            tx_errcnt = can_data[4]
            vdc = can_data[5] / 10
            l_zas = can_data[6]
            temperature = can_data[7]
            _LOG.debug("UPDATE DETAILS 0x02: {:08x} com_stat={} rx_errcnt={} tx_errcnt={} vdc={} l_zas={} temperature={}".format(
                self.can_id, com_stat, rx_errcnt, tx_errcnt, vdc, l_zas, temperature))
            self._details_flag |= 0x04

        elif data_id == 0x03:
            ecan_status = can_data[2]
            conditions_status = can_data[3]
            cpu_load = can_data[4]
            conditions = can_data[5] << 8
            _LOG.debug("UPDATE DETAILS 0x03: {:08x} ecan_status={} conditions_status={} cpu_load={} conditions={}".format(
                self.can_id, ecan_status, conditions_status, cpu_load, conditions))
            self._details_flag |= 0x08

        elif data_id == 0x04:
            time_quartz_errors = (can_data[2] << 16) | (can_data[3] << 8) | can_data[4]
            quarc_counter = (can_data[5] << 8) | can_data[6]
            _LOG.debug("UPDATE DETAILS 0x04: {:08x} time_quartz_errors={:06x} quarc_counter={:04x}".format(
                self.can_id, time_quartz_errors, quarc_counter))
            self._details_flag |= 0x10

        elif data_id == 0x05:
            new_can_id = int.from_bytes(can_data[2:], byteorder='little', signed=False)
            self.physical_can_id = self.can_id
            self.can_id = new_can_id
            update_can_id_callback(self, self.physical_can_id, new_can_id)
            self._details_flag |= 0x20

        elif data_id == 0x06:
            bin_inputs = can_data[2]
            bin_outputs = can_data[3]
            inputs = can_data[4]
            outputs = can_data[5]
            flags = can_data[6]
            line_flags = can_data[7]
            _LOG.debug(
                "UPDATE DETAILS 0x06: {:08x} bin_inputs={} bin_outputs={} inputs={} outputs={} flags={} line_flags={}".format(
                    self.can_id, bin_inputs, bin_outputs, inputs, outputs, flags, line_flags))
            self._details_flag |= 0x40
        elif data_id == 0x07:
            onewires = can_data[2]
            sub_procs = can_data[3]
            sub_type = can_data[4]
            mask = can_data[5]
            _LOG.debug("UPDATE DETAILS 0x07: {:08x} onewires={} sub_cpus={} sub_type={} mask={}".format(
                self.can_id, onewires, sub_procs, sub_type, mask))
            self._details_flag |= 0x80

    def state_changed(self, attribute, index):
        """Fire the state changed callback."""
        callbacks = self._state_update_callbacks.get((attribute, index))
        if callbacks:
            for callback in callbacks:
                callback()

    def register_state_changed_callback(self, attribute, index, callback):
        """Register calback for attribute state change."""
        self._state_update_callbacks[(attribute, index)].append(callback)
        return

    # TODO: Remove
    def update_listeners(self, attribute, index, old_value, new_value, unit=""):
        """Fire the callback for attribute change listeners."""
        callbacks = self._callbacks.get((attribute, index))
        if callbacks:
            for callback in callbacks:
                callback(self.can_id, attribute, index, old_value, new_value, unit)

    def add_listener(self, attribute, index, callback):
        """Add the callback for attribute value change event."""
        _LOG.debug("MAC: {} {} {} callback added".format(self.can_id, attribute, index))
        self._callbacks[(attribute, index)].append(callback)
        return
        # TODO: check if callback is callable
        # if attribute is None:
        #     attributes = self.attributes
        # else:
        #     attributes = {attribute: self.attributes.get(attribute, 0)}
        # for attribute, idx in attributes.items():
        #     if index is None:
        #         indices = range(1, idx + 1)
        #     else:
        #         indices = [index]
        #     for idx in indices:
        #         _LOG.debug("MAC: {} {} {} callback added".format(self.can_id, attribute, idx))
        #         self._callbacks[(attribute, idx)].append(callback)

    def remove_listener(self, attribute, index, callback):
        """Remove the callback."""
        callbacks = self._callbacks.get((attribute, index))
        if callbacks:
            try:
                callbacks.remove(callback)
            except ValueError:
                pass

    def on_change(self, broadcast_type, data):
        """Handle the on change event when attribute has changed."""
        # TODO: use map
        _LOG.debug("ON CHANGE: {:08x} {}: {}".format(self.can_id, broadcast_type, data))
        details = self.attributes.get(broadcast_type.value)
        if details:
            # [{'base': 1, 'name': 'temperature', 'max': 1, 'unit': 'C', 'converter': convert_temperature, 'min': 1}]
            for properties in details:
                name = properties['name']
                base = properties['base']
                max_index = properties['max']
                min_index = properties['min']
                unit = properties['unit']
                converter = properties['converter']
                index, old_value, new_value = data
                converted_index = index - min_index + base
                if min_index <= index <= max_index:
                    if converter is not None:
                        old_value = converter(old_value)
                        new_value = converter(new_value)
                    _LOG.debug("Changed {}({}) {}{}->{}{}".format(name, converted_index, old_value, unit, new_value, unit))
                    # remove
                    self.update_listeners(name, converted_index, old_value=old_value, new_value=new_value, unit=unit)
                    self.state_changed(attribute=name, index=converted_index)

    def get_attributes(self):
        """Generate the attribute names for module."""
        for _, attributes in self.attributes.items():
            for attribute_detail in attributes:
                name = attribute_detail['name']
                base = attribute_detail['base']
                min_broadcast_index = attribute_detail['min']
                max_broadcast_index = attribute_detail['max']
                unit = attribute_detail['unit']
                # TODO: simplify
                for index in range(min_broadcast_index, max_broadcast_index + 1):
                    conv_index = index - min_broadcast_index + base
                    yield name, conv_index, unit

    def _make_attribute_map(self):
        for broadcast_type, broadcast_details in self.attributes.items():
            for broadcast_detail in broadcast_details:
                attribute = broadcast_detail['name']
                base = broadcast_detail['base']
                min_broadcast_index = broadcast_detail['min']
                max_broadcast_index = broadcast_detail['max']
                unit = broadcast_detail['unit']
                converter = broadcast_detail['converter']
                for index in range(min_broadcast_index, max_broadcast_index + 1):
                    absolute_index = index - min_broadcast_index + base
                    self.map[(attribute, absolute_index)] = (BroadcastTypes(broadcast_type), index, unit, converter)

    def get_state(self, attribute, index):
        """Return the item state."""
        broadcast_type, index, unit, converter = self.map.get((attribute, index), (None, None, None, None))
        if broadcast_type is not None:
            value = self._broadcast_cache.get_value(broadcast_type, index)
            if value is not None:
                value = converter(value)
            return value
        else:
            return None

    def get_last_state(self, attribute, index):
        """Return the item last state."""
        broadcast_type, index, unit, converter = self.map.get((attribute, index), (None, None, None, None))
        if broadcast_type is not None:
            value = self._broadcast_cache.get_last_value(broadcast_type, index)
            if value is not None:
                value = converter(value)
            return value
        else:
            return None

    def get_measurement_unit(self, attribute, index):
        """Return measurement unit."""
        _, _, unit, _ = self.map.get((attribute, index), (None, None, None, None))
        return unit


def load_yaml(file_path):
    """Load YAML file from full file path and return dict."""
    with open(file_path, 'r') as yamlfile:
        try:
            dictionary = yaml.load(yamlfile)
        except yaml.YAMLError:
            raise
            return {}
    return dictionary


module_data = load_yaml(__file__[:-2] + "yaml")


def convert_unchanged(value):
    """Fake data converter."""
    return value


def validate_converter(value):
    """Validate the converter function."""
    try:
        return eval(value)
    except AttributeError:
        raise Invalid("Unknown converter function: {}".format(value))
    return None


def update_converter_if_none(value):
    """Make a default converter to transparently pass the value without conversion."""
    if value['converter'] is None:
        value['converter'] = convert_unchanged
    return value


def validate_broadcast(value):
    """Validate if broadcast name is known in module.yaml file."""
    new_value = {}
    for broadcast_name, details in value.items():
        try:
            broadcast_value = eval("BroadcastTypes." + broadcast_name).value
        except AttributeError:
            raise Invalid("Unknown Broadcast Name: {}".format(broadcast_name))
        new_value[broadcast_value] = details
    return new_value


BROADCAST_SCHEMA = All(Schema({
    Required('name'): str,
    Required('base'): int,
    Required('min'): int,
    Required('max'): int,
    Optional('unit', default=""): str,
    Required('converter', default=None): validate_converter,
}), update_converter_if_none)

ATTRIBUTE_SCHEMA = All(Schema({
    str: [BROADCAST_SCHEMA]
}), validate_broadcast)

MODULE_INFO_SCHEMA = All(Schema({
    Required('code'): Coerce(int),
    Required('protocol'): All(Lower, Any('any', Coerce(int))),
    Required('software_versions'): All([Coerce(int)]),
    Required('description'): Coerce(str),
    Required('attributes'): ATTRIBUTE_SCHEMA,
}))


SCHEMA = All(Schema({
    str: MODULE_INFO_SCHEMA,
}))


module_info = SCHEMA(module_data)


class ModuleFactory:
    """This is a ModuleFactory class to create specific module class based on discovered data."""

    factories = {}

    @staticmethod
    def create_module(can_id, code, protocol, software_version):
        """Create module object based on proivded data.

        Args:
            can_id (int): CAN ID
            code (int): Module Code Number
            protocol (int): Module CAN protocol version
            software_version (int): Module software version

        """
        max_match = 0
        matched_pn = None
        for part_number, details in module_info.items():
            match = 0
            if details['code'] == code:
                match += 1
            if (details['protocol'] in [protocol, "any"]) and match == 1:
                match += 1
            if software_version in details['software_versions'] and match == 2:
                match += 1
            if match > max_match:
                max_match = match
                matched_pn = part_number
            if max_match == 3:
                break

        if matched_pn is None:
            return None
        info = module_info.get(matched_pn)
        attributes = info.get('attributes')
        description = info.get("description")

        return AmpioModule(can_id=can_id, code=code, protocol=protocol, software_version=software_version,
                           part_number=matched_pn, description=description, attributes=attributes)


class AmpioModules:
    """This is a ModuleManager class implementation."""

    def __init__(self):
        """Initialize the :class: pyampio.modules.ModuleManager object."""
        self._modules = {}
        self._can_id_map_to_module = {}

    def __len__(self):
        """Return the number of modules."""
        return len(self._modules)

    def __iter__(self):
        """Iterate over module dictionary."""
        return iter(dict(self._modules).items())

    def add_module(self, can_id, can_data):
        """Add new module.

        Args:
            can_id (int): CAN ID
            can_data (bytearray): CAN frame data

        """
        sw_version = int((can_data[2] << 8) | can_data[3])
        pcb = int(can_data[4])
        code = int(can_data[1])
        protocol_version = int(can_data[5])
        mod = ModuleFactory.create_module(can_id=can_id, code=code, protocol=protocol_version, software_version=sw_version)
        try:
            mod.update_pcb_version(pcb)
        except AttributeError:
            _LOG.error("MODULE NOT IMPLEMENTED: {:08x} code={} {}".format(can_id, code, can_data))
            return

        self._modules[can_id] = mod
        self._can_id_map_to_module[can_id] = mod

    def update_name(self, can_id, can_data):
        """Update the module name.

        Args:
            can_id (int): CAN ID
            can_data (bytearray): CAN frame data

        """
        if can_data[0] == 0x0f and len(can_data) == 8:
            index = int(can_data[1])
            mod = self._modules.get(can_id)
            if mod is not None:
                mod.update_name(index, can_data[2:])
            else:
                _LOG.warning("Update name to unknown module: {:08x}".format(can_id))

    def update_details(self, can_id, can_data):
        """Update the module details.

        Args:
            can_id (int): CAN ID
            can_data (bytearray): CAN frame data

        """
        mod = self._modules.get(can_id)
        if mod is not None:
            mod.update_details(can_data, self.on_can_id_update)
        else:
            _LOG.warning("Update details to unknown module: {:08x}".format(can_id))

    def is_name_updated(self, can_id):
        """Return true if module name is updated else yield."""
        mod = self._modules.get(can_id)
        if mod:
            while not mod.is_name_complete:
                yield
            return True
        else:
            _LOG.warning("No module id {}".format(can_id))

    def is_details_updated(self, can_id):
        """Return true if module details are updated else yield."""
        start_time = dt.datetime.now()
        mod = self._modules.get(can_id)
        if mod:
            while not mod.is_details_complete and \
                    (dt.datetime.now() - start_time).seconds < QUERY_MODULE_DETAILS_TIMEOUT_SECONDS:
                yield
        if (dt.datetime.now() - start_time).seconds >= QUERY_MODULE_DETAILS_TIMEOUT_SECONDS:
            _LOG.warning("Query took more than {} seconds: {}".format(
                QUERY_MODULE_DETAILS_TIMEOUT_SECONDS, mod))
        return True

    def is_attribute_names_updated(self, can_id):
        """Return true if attribute name are updated else yield."""
        # TODO: Implementation needed
        while True:
            yield

    def on_can_id_update(self, mod, can_id, new_can_id):
        """Handle the CAN_ID update (i.e. physical -> logical)."""
        if can_id != new_can_id:
            self._can_id_map_to_module[new_can_id] = mod

    def broadcast_received(self, can_id, can_data):
        """Handle the received broadcast."""
        mod = self.get_module(can_id)
        if mod is None:
            _LOG.debug("Update for unknown module can_id={:08x}".format(can_id))
            return
        mod.broadcast_received(can_data)

    def add_on_value_changed_callback(self, can_id, attribute, index, callback):
        # TODO: Remove
        """Add the on value changed callback."""
        mod = self.get_module(can_id)
        if mod:
            for (attrib, idx, unit) in mod.get_attributes():
                if attribute is not None:
                    if attribute != attrib:
                        continue
                if index is not None:
                    if index != idx:
                        continue
                mod.add_listener(attrib, idx, partial(callback, self))
                _LOG.debug("On value changed callback added {} {} {}".format(can_id, attrib, idx))
        else:
            _LOG.warning("Module not known: {:08x}".format(can_id))

    def register_on_value_changed_callback(self, can_id, attribute, index, callback):
        """Register on value change_callback."""
        mod = self.get_module(can_id)
        attribute_re = re.compile(attribute)
        if mod:
            for (attrib, idx, unit) in mod.get_attributes():
                if attribute is not None:
                    if attribute_re.match(attrib) is None:
                        continue
                if index is not None:
                    if index != idx:
                        continue
                mod.register_state_changed_callback(attrib, idx, callback)
                _LOG.debug("On value changed callback added {} {} {}".format(can_id, attrib, idx))
        else:
            _LOG.warning("Module not known: {:08x}".format(can_id))

    @property
    def modules(self):
        """Return the module dictionary."""
        return self._modules

    def get_module(self, can_id):
        """Get the module from CAN ID."""
        return self._can_id_map_to_module.get(can_id)

    def get_attributes(self, can_id, attribute):
        """Generate the module attributes."""
        mod = self.get_module(can_id)
        if mod:
            for (attr, index, _) in mod.get_attributes():
                if attr == attribute:
                    yield attribute, index

    def get_state(self, can_id, attribute, index):
        """Return item state."""
        mod = self.get_module(can_id)
        if mod:
            return mod.get_state(attribute, index)
        else:
            return None

    def get_last_state(self, can_id, attribute, index):
        """Return item last state."""
        mod = self.get_module(can_id)
        if mod:
            return mod.get_last_state(attribute, index)
        else:
            return None

    def get_measurement_unit(self, can_id, attribute, index):
        """Return measurement unit."""
        mod = self.get_module(can_id)
        if mod:
            return mod.get_measurement_unit(attribute, index)
        else:
            return None

    def get_module_name(self, can_id):
        """Return module name."""
        mod = self.get_module(can_id)
        if mod:
            return mod.name
        else:
            return None

    def get_module_part_number(self, can_id):
        """Return module name."""
        mod = self.get_module(can_id)
        if mod:
            return mod.part_number
        else:
            return None



"""This is a broadcast type definition module."""

from enum import Enum
import logging

_LOG = logging.getLogger(__name__)


def bits_to_indices(mask):
    """Generate list of indices of bits set to one in the value."""
    i = 1
    while mask > 0:
        if bool(mask & 0x1):
            yield i
        mask >>= 1
        i += 1
    return


class BroadcastTypes(Enum):
    """The BroadcastTypes enum with known broadcast frame types."""

    TemperatureInt = 0x05
    Temperature1T3 = 0x06
    Temperature4T6 = 0x07
    Analog1T6 = 0x0c
    Analog7T12 = 0x0d
    Analog13T18 = 0x0e
    BinInOut = 0x0f
    DateTime = 0x10
    Value32b = 0x18
    SatelArmed = 0x19
    SatelAlarm = 0x1a
    SatelInput1T48 = 0x1b
    SatelInput49T96 = 0x1c
    SatelInput97T128 = 0x1d
    SatelOutput1T48 = 0x1e
    SatelOutput49T96 = 0x1f
    SatelOutput97T128 = 0x20
    Value16b1T3 = 0x21
    Value16b4T6 = 0x22
    Value16b7T9 = 0x23
    Value16b10T12 = 0x24
    Value16b13T15 = 0x25
    Value8b1T6 = 0x26
    Value8b7T12 = 0x27
    Value8b13T18 = 0x28
    Event = 0x2b
    SatelInput129T176 = 0x31
    SatelInput177T224 = 0x32
    SatelInput225T256 = 0x33
    SatelOutput129T176 = 0x34
    SatelOutput177T224 = 0x35
    SatelOutput225T256 = 0x36
    SatelBreached = 0x38
    SatelArming = 0x39
    SatelArming10s = 0x3a
    HeatingZoneSummary = 0xc8
    HeatingZone1 = 0xc9
    HeatingZone2 = 0xca
    HeatingZone3 = 0xcb
    HeatingZone4 = 0xcc
    HeatingZone5 = 0xcd
    HeatingZone6 = 0xce
    HeatingZone7 = 0xcf
    HeatingZone8 = 0xd0
    HeatingZone9 = 0xd1
    HeatingZone10 = 0xd2
    HeatingZone11 = 0xd3
    HeatingZone12 = 0xd4
    HeatingZone13 = 0xd5
    HeatingZone14 = 0xd6
    HeatingZone15 = 0xd7
    HeatingZone16 = 0xd8
    Flag = 0x80


broadcasts = {}


def register_broadcast(broadcast_types, target_class):
    """Register broadcast class.."""
    for broadcast_type in broadcast_types:
        broadcasts[broadcast_type] = target_class


class BroadcastMeta(type):
    """This is BroadcastMeta class for Broadcast class self registration."""

    def __new__(cls, clsname, bases, attrs):
        """Register the broadcast class."""
        newclass = super(cls, BroadcastMeta).__new__(cls, clsname, bases, attrs)
        register_broadcast(newclass.broadcast_types, newclass)
        return newclass


class BroadcastFactory:
    """This is a BroadcastFactory class for Broadcast object creation."""

    @staticmethod
    def create_broadcast(data):
        """Create the Broadcast object based on the frame date."""
        try:
            broadcast_type = BroadcastTypes(data[0])
        except ValueError:
            _LOG.debug("Unknown broadcast type: {:02x}".format(data[0]))
            return None

        return broadcasts.get(broadcast_type)(data[1:])


class BroadcastCache:
    """The BroadcastCache object is a cache for received broadcast for module."""

    def __init__(self, on_changed=None):
        """Initialize the BroadcastCache object."""
        self._cache = {}
        self._on_changed_callback = on_changed

    def update(self, data):
        """Update the cache with broadcast data."""
        try:
            broadcast_type = BroadcastTypes(data[0])
        except ValueError:
            _LOG.debug("Update for unknown broadcast type: {:02x}".format(data[0]))
            return

        if broadcast_type in self._cache:
            self._cache[broadcast_type].update(data[1:])
        else:
            broadcast = BroadcastFactory.create_broadcast(data)
            if broadcast is not None:
                self._cache[broadcast_type] = broadcast
            else:
                _LOG.warning("Unknown broadcast: {}:{}".format(broadcast, broadcast_type))
                return
        try:
            for change in self._cache[broadcast_type].changes():
                _LOG.debug("Type: {} Change: {}".format(type(self._cache[broadcast_type]), change))
                if self._on_changed_callback is not None:
                    self._on_changed_callback(broadcast_type, change)
        except TypeError:
            _LOG.critical("Data: {}".format(data))
            _LOG.critical("Broadcast type: {}".format(broadcast_type))
            _LOG.critical("Cache: {}".format(self._cache))
            raise

    def get_value(self, broadcast_type, index):
        """Get the value state from broadcast.

        Returns: value state or None if not received yet

        """
        broadcast_object = self._cache.get(broadcast_type, None)
        if broadcast_object:
            return broadcast_object.state(index)
        else:
            return None

    def get_last_value(self, broadcast_type, index):
        """Get the value state from broadcast.

        Returns: value state or None if not received yet

        """
        broadcast_object = self._cache.get(broadcast_type, None)
        if broadcast_object:
            return broadcast_object.last_state(index)
        else:
            return None


class BroadcastBinary(metaclass=BroadcastMeta):
    """This is a BroadcastBinary class representing binary data frames."""

    broadcast_types = [
        BroadcastTypes.Flag,
        BroadcastTypes.BinInOut,
        BroadcastTypes.HeatingZoneSummary,
        BroadcastTypes.SatelInput1T48,
        BroadcastTypes.SatelInput49T96,
        BroadcastTypes.SatelInput97T128,
        BroadcastTypes.SatelInput129T176,
        BroadcastTypes.SatelInput177T224,
        BroadcastTypes.SatelInput225T256,
        BroadcastTypes.SatelOutput1T48,
        BroadcastTypes.SatelOutput49T96,
        BroadcastTypes.SatelOutput97T128,
        BroadcastTypes.SatelOutput129T176,
        BroadcastTypes.SatelOutput177T224,
        BroadcastTypes.SatelOutput225T256,
        BroadcastTypes.SatelArmed,
        BroadcastTypes.SatelAlarm,
        BroadcastTypes.SatelBreached,
        BroadcastTypes.SatelArming,
        BroadcastTypes.SatelArming10s,
    ]

    def __init__(self, data=None):
        """Initialize BroadcastBinary class from frame data."""
        self._value = int.from_bytes(data[2:], byteorder='little', signed=False) if data else 0
        self._previous_value = 0

    def update(self, data):
        """Update BroadcastBinary class from frame data."""
        value = int.from_bytes(data, byteorder='little', signed=False)
        self._previous_value = self._value
        self._value = value

    def changes(self):
        """Yield the value changes in broadcast data."""
        mask = self._previous_value ^ self._value
        for index in bits_to_indices(mask):
            previous_value = bool(self._previous_value & (0x1 << (index - 1)))
            new_value = bool(self._value & (0x1 << (index - 1)))
            yield index, previous_value, new_value

    def state(self, index):
        """Return the current value state from broadcast."""
        return bool(self._value & (0x1 << (index - 1)))

    def last_state(self, index):
        """Return the last value state from broadcast."""
        return bool(self._previous_value & (0x1 << (index - 1)))


class BroadcastValue8b(metaclass=BroadcastMeta):
    """This is a BroadcastValue8b class representing 8-bit data frames."""

    broadcast_types = [
        BroadcastTypes.Analog1T6,
        BroadcastTypes.Analog7T12,
        BroadcastTypes.Analog13T18,
        BroadcastTypes.TemperatureInt,
        BroadcastTypes.Value8b1T6,
        BroadcastTypes.Value8b7T12,
        BroadcastTypes.Value8b13T18,
    ]

    def __init__(self, data=None):
        """Initialize BroadcastValue8b class from frame data."""
        self._values = [None] * 6
        self._previous_values = [None] * 6
        if data:
            self.update(data)

    def update(self, data):
        """Update BroadcastValue8b class from frame data."""
        self._previous_values = self._values
        self._values = [int(temp) for temp in data] if data else [0] * 6
        _LOG.debug("Value 8b: Value {}".format(self._values))

    def changes(self):
        """Yield the value changes in broadcast data."""
        for index, (current, previous) in enumerate(zip(self._values, self._previous_values), start=1):
            if current != previous:
                yield index, previous, current

    def state(self, index):
        """Return the current value state from broadcast."""
        try:
            return self._values[index - 1]
        except IndexError:
            return None
        return None

    def last_state(self, index):
        """Return the last value state from broadcast."""
        try:
            return self._previous_values[index - 1]
        except IndexError:
            return None
        return None


class BroadcastValue16b(metaclass=BroadcastMeta):
    """This is a BroadcastValue16b class representing 16-bit data frames."""

    broadcast_types = [
        BroadcastTypes.Value16b1T3,
        BroadcastTypes.Value16b4T6,
        BroadcastTypes.Value16b7T9,
        BroadcastTypes.Value16b10T12,
        BroadcastTypes.Value16b13T15,
        BroadcastTypes.Temperature1T3,
        BroadcastTypes.Temperature4T6,
    ]

    def __init__(self, data=None):
        """Initialize BroadcastValue16b class from frame data."""
        self._values = [0] * 3
        self._previous_values = [0] * 3
        if data:
            self.update(data)

    def update(self, data):
        """Update BroadcastValue16b class from frame data."""
        self._previous_values = self._values
        self._values = [int.from_bytes(temp, byteorder='little', signed=False)
                        for temp in zip(*[iter(data)] * 2)] if data else [0] * 3
        _LOG.debug("Value 16b: Value {}".format(self._values))

    def state(self, index):
        """Return the current value state from broadcast."""
        try:
            return self._values[index - 1]
        except IndexError:
            return None
        return None

    def last_state(self, index):
        """Return the last value state from broadcast."""
        try:
            return self._previous_values[index - 1]
        except IndexError:
            return None
        return None

    def changes(self):
        """Yield the value changes in broadcast data."""
        for index, (current, previous) in enumerate(zip(self._values, self._previous_values), start=1):
            if current != previous:
                yield index, previous, current


class BroadcastValue32b(metaclass=BroadcastMeta):
    """This is a BroadcastValue32b class representing 32-bit data frames."""

    broadcast_types = [
        BroadcastTypes.Value32b
    ]

    def __init__(self, data=None):
        """Initialize BroadcastValue32b class from frame data."""
        self._values = {}
        self._previous_values = {}
        if data:
            self.update(data)

    def update(self, data):
        """Update BroadcastValue32b class from frame data."""
        index = int.from_bytes(data[0:2], byteorder="big", signed=False)
        value = int.from_bytes(data[2:], byteorder="little", signed=False)
        self._previous_values[index] = self._values.get(index, 0)
        self._values[index] = value
        _LOG.debug("Value 32b: Index {} Value {}".format(index + 1, value))

    def state(self, index):
        """Return the current value state from broadcast."""
        try:
            return self._values[index - 1]
        except IndexError:
            return None
        return None

    def last_state(self, index):
        """Return the last value state from broadcast."""
        try:
            return self._previous_values[index - 1]
        except IndexError:
            return None
        return None

    def changes(self):
        """Yield the value changes in broadcast data."""
        for index in self._values.keys():
            if self._previous_values[index] != self._values[index]:
                yield index + 1, self._previous_values[index], self._values[index]


class BroadcastHeatingZone(metaclass=BroadcastMeta):
    """This is a BroadcastHeatingZone class representing heating zone information."""

    broadcast_types = [
        BroadcastTypes.HeatingZone1,
        BroadcastTypes.HeatingZone2,
        BroadcastTypes.HeatingZone3,
        BroadcastTypes.HeatingZone4,
        BroadcastTypes.HeatingZone5,
        BroadcastTypes.HeatingZone6,
        BroadcastTypes.HeatingZone7,
        BroadcastTypes.HeatingZone8,
        BroadcastTypes.HeatingZone9,
        BroadcastTypes.HeatingZone10,
        BroadcastTypes.HeatingZone11,
        BroadcastTypes.HeatingZone12,
        BroadcastTypes.HeatingZone13,
        BroadcastTypes.HeatingZone14,
        BroadcastTypes.HeatingZone15,
        BroadcastTypes.HeatingZone16,

    ]

    def __init__(self, data=None):
        """Initialize BroadcastHeatingZone class from frame data."""
        self._values = [0] * 7
        self._previous_values = [0] * 7
        if data:
            self.update(data)

    def update(self, data):
        """Update BroadcastHeatingZone class from frame data."""
        self._previous_values = list(self._values)
        # temp measured
        self._values[0] = int.from_bytes(data[:2], byteorder='little', signed=False)
        # temp setpoint
        self._values[1] = int.from_bytes(data[2:4], byteorder='little', signed=False)
        # control mode
        self._values[2] = int(data[4])

        zone_params = int(data[5])
        # active
        self._values[3] = bool(zone_params & 0x01)
        # heating
        self._values[4] = bool(zone_params & 0x02)
        # day_mode
        self._values[5] = bool(zone_params & 0x04)
        # mode
        self._values[6] = int(zone_params & 0x70)
        _LOG.debug("Heating Zone Update: {}".format(self._values))

    def state(self, index):
        """Return the current value state from broadcast."""
        try:
            return self._values[index - 1]
        except IndexError:
            return None
        return None

    def last_state(self, index):
        """Return the last value state from broadcast."""
        try:
            return self._previous_values[index - 1]
        except IndexError:
            return None
        return None

    def changes(self):
        """Yield the value changes in broadcast data."""
        for index, (current, previous) in enumerate(zip(self._values, self._previous_values), start=1):
            if current != previous:
                yield index, previous, current


class BroadcastDateTime(metaclass=BroadcastMeta):
    """This is BroadcastDateTime representing date/time information."""

    broadcast_types = [
        BroadcastTypes.DateTime
    ]

    def __init__(self, data=None):
        """Initialize BroadcastDateTime class from frame data."""
        self._values = [None] * 7
        self._previous_values = [None] * 7

        if data is not None:
            self.update(data)

    def update(self, data):
        """Update BroadcastDateTime class from frame data."""
        self._previous_values = list(self._values)
        # year
        self._values[0] = int(data[0])
        # months
        self._values[1] = int(data[1] & 0x0f)
        # day
        self._values[2] = int(data[2] & 0x1f)
        # weekday
        self._values[3] = int(data[3] & 0x07)
        # daytime
        self._values[4] = bool(data[4] & 0x80)
        # hour
        self._values[5] = int(data[4] & 0x1f)
        # minute
        self._values[6] = int(data[5] & 0x7f)
        _LOG.debug("20{}-{}-{} {}:{} Weekday={} Day={}".format(
            self._values[0], self._values[1], self._values[2],
            self._values[5], self._values[6], self._values[3], self._values[4]))

    def state(self, index):
        """Return the current value state from broadcast."""
        try:
            return self._values[index - 1]
        except IndexError:
            return None
        return None

    def last_state(self, index):
        """Return the last value state from broadcast."""
        try:
            return self._previous_values[index - 1]
        except IndexError:
            return None
        return None

    def changes(self):
        """Yield the value changes in broadcast data."""
        for index, (current, previous) in enumerate(zip(self._values, self._previous_values), start=1):
            if current != previous:
                yield index, previous, current


class BroadcastEvent(metaclass=BroadcastMeta):
    """This is a BroadcastEvent frame representing the Ampio Event frame."""

    broadcast_types = [
        BroadcastTypes.Event
    ]

    def __init__(self, data=None):
        """Initialize BroadcastEvent class from frame data."""
        self._previous_values = [False] * 256
        self._values = [False] * 256
        if data is not None:
            self.update(data)
        pass

    def update(self, data):
        """Update BroadcastEvent class from frame data."""
        index = int(data[0])
        self._values[index] = True
        self._previous_values[index] = False

    def state(self, index):
        """Return the current value state from broadcast."""
        try:
            return self._values[index]
        except IndexError:
            return None
        return None

    def last_state(self, index):
        """Return the last value state from broadcast."""
        try:
            return self._previous_values[index - 1]
        except IndexError:
            return None
        return None

    def changes(self):
        """Yield the value changes in broadcast data."""
        for index, (current, previous) in enumerate(zip(self._values, self._previous_values), start=1):
            if current != previous:
                yield index, previous, current
                self._values[index] = self._previous_values[index] = False
